/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.ow2.opensuit.xmlmap.mapping.ElementMapping;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.schema.ISchema;
import org.ow2.opensuit.xmlmap.schema.ISchemaElement;
import org.ow2.opensuit.xmlmap.schema.SchemasManager;
import org.ow2.opensuit.xmlmap.utils.ClassHelper;



/**
 * This class is a program (main) that allows generating the XML documentation
 */
public class DocGenerator
{
	private static final boolean SHOW_INHERITANCE_IN_FIELDS = false;
	
	private static final String[] STRING_ARRAY = {};
	private static final ISchemaElement[] ELT_ARRAY_TYPE = {};

	/**
     * The DocGenerator program.
     * <p>Usage:</p>
     * <pre>DocGenerator <font color=red>&lt;dest_doc_dir&gt;</font> <font color=red>&lt;java_package&gt;</font> [ <font color=red>&lt;import_package&gt;</font> ]*
     * <p>Exemple:</p>
     * <pre>DocGenerator <font color=red>target/xmldoc org.ow2.opensuit</font>
     * 
     * @param args The program arguments.
     */
	public static void main(String[] args)
	{
		// --- analyse arguments
		if(args.length < 2)
		{
			System.out.println("usage: DocGenerator [dest_doc_dir] [java_package] ( [import_package] )*");
			return;
		}
		File outDir = new File(args[0]);
		String javaPackage = args[1];
		try
		{
			System.out.println("Generate Documentation from code");
			System.out.println("--------------------------------");
			System.out.println("  - output directory : "+outDir);
			System.out.println("  - root code package: "+javaPackage);
			
			// --- create schemas
			SchemasManager schemas = new SchemasManager();
			ISchema mainSchema = schemas.declareSchema(javaPackage, javaPackage, javaPackage);
			for(int i=2; i<args.length; i++)
			{
				String importPack = args[i];
//				System.out.println("  --> import file : "+importXsd+", prefix : "+importPrefix+", Java package: "+importPack);
				schemas.declareSchema(importPack, importPack, importPack);
			}
			
			System.out.println("");
			
			// --- build main schema from source code
			mainSchema.buildFromClassPath();
			
			DocGenerator generator = new DocGenerator(schemas);
			generator.buildDoc(mainSchema , outDir.getAbsoluteFile());
			
			System.out.println("... Done.");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	private SchemasManager schemas;
	private ISchema mainSchema;
	private File docRootDir;
	private Locale locale = Locale.US;
	private String generatedComment;
	
	public DocGenerator(SchemasManager iSchemas)
	{
		schemas = iSchemas;
	}
	public void buildDoc(ISchema iSchema, File iDocRootDir) throws FileNotFoundException
	{
		mainSchema = iSchema;
		docRootDir = iDocRootDir.getAbsoluteFile();
		
		// --- clean destination directory (recursively)
		//recurDeleteDir(docRootDir);
		
		// --- generation comment
		Date now = new Date();
		generatedComment = "Generated by XMLMap DocGenerator on "+now;
		
		// --- extract info
		ISchemaElement[] elts = mainSchema.getAllElements();
		List<String> tmp = new ArrayList<String>();
		for(int i=0; i<elts.length; i++)
		{
			String pack = elts[i].getTagName();
			int idx = pack.lastIndexOf('.');
			
			if(idx > 0){
				pack = pack.substring(0, idx);
			} else{
				pack = "";
			}
			
			if(!tmp.contains(pack)){
				tmp.add(pack);
			}
		}
		String[] packages = (String[])tmp.toArray(STRING_ARRAY);
		// --- sort
		Arrays.sort(packages);
		
		// --- 1: frameset page (index.html)
		writeFrameset();
		
		// --- 2: the packages frame
		writePackagesList(packages);
		
		// --- 3: all classes frame
		writeClassesList(null);
		
		// --- 4: classes from package frame(s)
		for(int i=0; i<packages.length; i++){
			writeClassesList(packages[i]);
		}
		
		// --- package summary pages
		writeSchemaSummary(iSchema);
		
		// --- elements pages
		for(int i=0; i<elts.length; i++){
			writeElement(elts[i]);
		}
	}
	/*
	private static boolean recurDeleteDir(File iDir)
	{
		if(!iDir.exists())
			return true;
		File[] files = iDir.listFiles();
		for(int i=0; i<files.length; i++)
		{
			if(files[i].isDirectory())
				recurDeleteDir(files[i]);
			else
				files[i].delete();
		}
		iDir.delete();
		return true;
	}
	*/
	/**
	 * path is relative to the root doc directory
	 */
	private File newFile(String path)
	{
		return new File(docRootDir.getAbsolutePath()+"/"+path);
	}

	private void writeFrameset() throws FileNotFoundException
	{
		File f = newFile(mainSchema.getRootPackage()+"_index.html");
		f.getParentFile().mkdirs();
		PrintWriter writer = new PrintWriter(new FileOutputStream(f));
		
		writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" \"http://www.w3.org/TR/html4/frameset.dtd\">");
		writer.println("<HTML>");
		
		writer.println("<HEAD>");
		writer.println("<!-- "+generatedComment+"  -->");
		writer.println("<TITLE>");
		writer.println("XML Doc for "+mainSchema.getName());
		writer.println("</TITLE>");
		writer.println("</HEAD>");
		
		writer.println("<FRAMESET cols=\"20%,80%\">");
		writer.println("	<FRAMESET rows=\"30%,70%\">");
		writer.println("		<FRAME src=\""+mainSchema.getRootPackage()+"/overview-frame.html\" name=\"packageListFrame\" title=\"All Packages\">");
		writer.println("		<FRAME src=\""+mainSchema.getRootPackage()+"/allclasses-frame.html\" name=\"packageFrame\" title=\"All elements and substitution groups\">");
		writer.println("	</FRAMESET>");
		writer.println("	<FRAME src=\""+mainSchema.getRootPackage()+"/overview-summary.html\" name=\"classFrame\" title=\"Package, element and substitution group descriptions\">");
		writer.println("</FRAMESET>");
		writer.println("</HTML>");
		
		writer.close();
	}
	private void writePackagesList(String[] packages) throws FileNotFoundException
	{
		File f = newFile(mainSchema.getRootPackage()+"/overview-frame.html");
		f.getParentFile().mkdirs();
		PrintWriter writer = new PrintWriter(new FileOutputStream(f));
	
		writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
		writer.println("<HTML>");
		writer.println("<HEAD>");
		writer.println("<!-- "+generatedComment+" -->");
		writer.println("<TITLE>");
		writer.println("Packages List");
		writer.println("</TITLE>");
		writer.println("<LINK REL =\"stylesheet\" TYPE=\"text/css\" HREF=\"stylesheet.css\" TITLE=\"Style\">");
		writer.println("<base target=\"packageFrame\">");
		writer.println("</HEAD>");
		
		writer.println("<BODY BGCOLOR=\"white\">");
		/*
		writer.println("<TABLE BORDER=\"0\" WIDTH=\"100%\">");
		writer.println("<TR>");
		writer.println("<TD NOWRAP><FONT size=\"+1\" CLASS=\"FrameTitleFont\">");
		writer.println("<B><b>Java<sup><font size=-2>TM</font></sup> 2 Platform<br>Ent. Ed. v1.4</b></B></FONT></TD>");
		writer.println("</TR>");
		writer.println("</TABLE>");
		*/
		writer.println("<TABLE BORDER=\"0\" WIDTH=\"100%\">");
		writer.println("<TR>");
		writer.println("<TD NOWRAP><FONT CLASS=\"FrameItemFont\"><A HREF=\"allclasses-frame.html\">All Classes</A></FONT>");
		writer.println("<P>");
		writer.println("<FONT size=\"+1\" CLASS=\"FrameHeadingFont\">Packages</FONT>");
		writer.println("<BR>");
		
		// --- render links
		for(int i=0 ;i<packages.length; i++)
		{
			String packageName = packages[i];
			String packageDir = null;
			
			if(packageName.length() == 0)
			{
				packageName = "(root)";
				packageDir = ".";
			}
			else
			{
				packageDir = packages[i].replace('.', '/');
			}
			writer.println("<FONT CLASS=\"FrameItemFont\"><A HREF=\""+packageDir+"/package-frame.html\">"+packageName+"</A></FONT>");
			writer.println("<BR>");
		}
		
		writer.println("</TD>");
		writer.println("</TR>");
		writer.println("</TABLE>");
		writer.println("</BODY>");
		writer.println("</HTML>");
	
		writer.close();
	}
	private String toRoot(File file, File dir)
	{
		File d = file.getParentFile();
        final StringBuilder toRoot = new StringBuilder(".");
        while(d != null && !d.equals(dir))
        {
            toRoot.append("/..");
            d = d.getParentFile();
        }
        if(d == null)
        {
            System.out.println("error: toRoot("+file+", "+dir+")... null");
        }
        return toRoot.toString();
	}
	private void writeClassesList(String pack) throws FileNotFoundException
	{
		String toRoot = ".";
		String title = "All Elements";
		File f = null;
		if(pack == null)
		{
			f = newFile(mainSchema.getRootPackage()+"/allclasses-frame.html");
			toRoot = "..";
		}
		else
		{
			if(pack.length() == 0)
			{
				// root package
				title = "Elements from root package";
				f = newFile(mainSchema.getRootPackage()+"/package-frame.html");
				toRoot = "..";
			}
			else
			{
				// non-root package
				title = "Elements from package '"+pack+"'";
				f = newFile(mainSchema.getRootPackage()+"/"+(pack.replace('.', '/'))+"/package-frame.html");
				toRoot = toRoot(f, docRootDir);
			}
		}
		f.getParentFile().mkdirs();
		PrintWriter writer = new PrintWriter(new FileOutputStream(f));
	
		
		writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
		writer.println("<HTML>");
		writer.println("<HEAD>");
		writer.println("<!-- "+generatedComment+"  -->");
		writer.println("<TITLE>");
		writer.println(title);
		writer.println("</TITLE>");
		writer.println("<LINK REL =\"stylesheet\" TYPE=\"text/css\" HREF=\""+toRoot+"/stylesheet.css\" TITLE=\"Style\">");
		writer.println("<base target=\"classFrame\">");
		writer.println("</HEAD>");

		writer.println("<BODY BGCOLOR=\"white\">");
		writer.println("<FONT size=\"+1\" CLASS=\"FrameHeadingFont\">");
		writer.println("<B>"+title+"</B></FONT>");
		writer.println("<BR>");

		writer.println("<TABLE BORDER=\"0\" WIDTH=\"100%\">");
		writer.println("<TR>");
		writer.println("<TD NOWRAP><FONT CLASS=\"FrameItemFont\">");

		// --- get elements
		ISchemaElement[] elts = mainSchema.getAllElements();
		if(pack != null)
		{
			List<ISchemaElement> eltsFromPackage = new ArrayList<ISchemaElement>();
			for(int i=0; i<elts.length; i++)
			{
				String eltPack = elts[i].getTagName();
				int idx = eltPack.lastIndexOf('.');
				
				if(idx > 0){
					eltPack = eltPack.substring(0, idx);
				}else{
					eltPack = "";
				}
				
				if(eltPack.equals(pack)){
					eltsFromPackage.add(elts[i]);
				}
			}

			elts = (ISchemaElement[])eltsFromPackage.toArray(ELT_ARRAY_TYPE);
		}
		
		// --- sort
		Arrays.sort(elts, new Comparator(){
			public int compare(Object o1, Object o2)
			{
				ISchemaElement e1 = (ISchemaElement)o1;
				ISchemaElement e2 = (ISchemaElement)o2;
				return e1.getTagName().compareTo(e2.getTagName());
			}
		});
		
		// --- render links
		for(int i=0 ;i<elts.length; i++)
		{
			String link = getElementDocPath(elts[i]);
			
			if(elts[i].getMappedClass().isInterface()){
				writer.println("<I>");
			}
			
			writer.println("<A HREF=\""+toRoot+"/"+link+"\">"+elts[i].getTagName()+"</A>");
			
			if(elts[i].getMappedClass().isInterface()){
				writer.println("</I>");
			}
			
			writer.println("<BR>");
		}
		
		writer.println("</TD>");
		writer.println("</TR>");
		writer.println("</TABLE>");
		writer.println("</BODY>");
		writer.println("</HTML>");
	
		writer.close();
	}
	private void writeSchemaSummary(ISchema schema) throws FileNotFoundException
	{
		// --- create html file
		File f = newFile(mainSchema.getRootPackage()+"/overview-summary.html");
		f.getParentFile().mkdirs();
		String toRoot = toRoot(f, docRootDir);
		
		// --- render
		PrintWriter writer = new PrintWriter(new FileOutputStream(f));
		writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
		writer.println("<HTML>");
		writer.println("<HEAD>");
		writer.println("<!-- "+generatedComment+"  -->");
		writer.println("<TITLE>");
		writer.print("XML Doc: ");
		writer.print(schema.getRootPackage());
		writer.println("</TITLE>");
		writer.println("<LINK REL =\"stylesheet\" TYPE=\"text/css\" HREF=\""+toRoot+"/stylesheet.css\" TITLE=\"Style\">");
		writer.println("<base target=\"classFrame\">");
		writer.println("</HEAD>");
		
		writer.println("<BODY BGCOLOR=\"white\">");
		
		// --- title
		writer.println("<H2>");
		writer.print("XML Doc");
		writer.println("</H2>");
		
		writer.println("<HR>");
		
		// --- info
		writer.println("<P>");
		writer.print("This is the XML documentation for Java package <b>");
		writer.print(schema.getRootPackage());
		writer.print("</b>.");
		writer.println("</P>");
		
		writer.println("<HR>");
		
		// --- dependencies
		ISchema[] dependencies = schema.getUsedSchemas();
		
		writer.print("<h3>");
		writer.print("Schema Dependencies:");
		writer.println("<h3>");
		
		writer.println("<ul>");
		if(dependencies != null && dependencies.length > 0)
		{
			for(int i=0; i<dependencies.length; i++)
			{
				writer.print("<li>");
				writer.print("<a href=\"../");
				writer.print(dependencies[i].getRootPackage());
				writer.print("_index.html\" target=\"_top\">");
				writer.print(dependencies[i].getRootPackage());
				writer.print("</a>");
				writer.println("</li>");
			}
		}
		else
		{
			writer.println("<li>none</li>");
		}
	
		writer.println("</BODY>");
		writer.println("</HTML>");
	
		writer.close();
	}
	/**
	 * Returns the hyperlink to the given element XML doc.
	 * The url is relative to the root doc directory.
	 * @param iElt
	 * @return
	 */
	private String getElementDocPath(ISchemaElement iElt)
	{
		// --- allow link between different schemas
//		if(iElt.getSchema() != mainSchema)
//			return null;
		return iElt.getSchema().getRootPackage()+"/"+iElt.getTagName().replace('.', '/')+".html";
	}
	private String getPackageDocPath(String packageName)
	{
		if(packageName == null){
			return mainSchema.getRootPackage();
		}
		return mainSchema.getRootPackage()+"/"+(packageName.replace('.', '/'));
	}
	private void writeElement(ISchemaElement iElt) throws FileNotFoundException
	{
		// --- collect info
		boolean isAbstract = (iElt.getMappedClass().getModifiers() & Modifier.ABSTRACT) != 0;
		int idx = iElt.getTagName().lastIndexOf('.');
		String simpleEltName = iElt.getTagName();
		String packageName = "(root)";
		if(idx > 0)
		{
			packageName = iElt.getTagName().substring(0, idx);
			simpleEltName = iElt.getTagName().substring(idx+1);
		}
		String title = null;
		if(iElt.getType() == ISchemaElement.ELEMENT_TYPE)
		{
			if(isAbstract){
				title = "abstract Element "+iElt.getTagName();
			} else{
				title = "Element "+iElt.getTagName();
			}
		}
		else if(iElt.getType() == ISchemaElement.ENUMERATION_TYPE)
		{
			title = "Enumeration "+iElt.getTagName();
		}
		else
		{
			title = "Substitution Group "+iElt.getTagName();
		}
		
		// --- get ancestors
		List<ISchemaElement> ancestors = new ArrayList<ISchemaElement>();
		ISchemaElement ancestor = iElt.getSuperType();
		while(ancestor != null)
		{
			ancestors.add(0, ancestor);
			ancestor = ancestor.getSuperType();
		}
		ancestors.add(iElt);
		
		// --- get field ancestors
		List<ISchemaElement> fieldAncestors = new ArrayList<ISchemaElement>();
		List<ElementMapping> mappings = new ArrayList<ElementMapping>();
		if(SHOW_INHERITANCE_IN_FIELDS)
		{
			for(int i=0; i<ancestors.size(); i++)
			{
				fieldAncestors.add(ancestors.get(i));
				mappings.add(iElt.getSchema().getMethod().getElementMappings(((ISchemaElement)fieldAncestors.get(i)).getMappedClass(), false));
			}
		}
		else
		{
			fieldAncestors.add(iElt);
			mappings.add(iElt.getSchema().getMethod().getElementMappings(iElt.getMappedClass(), true));
		}
		
		// --- create html file
		File f = newFile(getElementDocPath(iElt));
		f.getParentFile().mkdirs();
		String toRoot = toRoot(f, docRootDir);
		
		// --- render
		PrintWriter writer = new PrintWriter(new FileOutputStream(f));
		writer.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">");
		writer.println("<HTML>");
		writer.println("<HEAD>");
		writer.println("<!-- "+generatedComment+"  -->");
		writer.println("<TITLE>");
		writer.println(title);
		writer.println("</TITLE>");
		writer.println("<LINK REL =\"stylesheet\" TYPE=\"text/css\" HREF=\""+toRoot+"/stylesheet.css\" TITLE=\"Style\">");
		writer.println("<base target=\"classFrame\">");
		writer.println("</HEAD>");
		
		writer.println("<SCRIPT type=\"text/javascript\">");
		writer.println("function windowTitle()");
		writer.println("{");
		writer.println("	parent.document.title=\""+title+"\";");
		writer.println("}");
		writer.println("</SCRIPT>");

		writer.println("<BODY onload=\"windowTitle()\" BGCOLOR=\"white\">");
		
		// --- title
		writer.println("<H2>");
		/*
		writer.println("<FONT SIZE=\"-1\">");
		writer.println(packageName+"</FONT>");
		writer.println("<BR>");
		writer.println("Class HttpServlet</H2>");
		*/
		writer.println(title);
		writer.println("</H2>");
		
		// --- inheritance graph
		if(ancestors.size() > 1)
		{
			writer.println("<PRE>");
			for(int i=0; i<ancestors.size(); i++)
			{
				if(i>0)
				{
					for(int j=0;j<i; j++){
						writer.print("  ");
					}
					
					writer.println("|");
					
					for(int j=0;j<i; j++){
						writer.print("  ");
					}
					
					writer.print("+-- ");
				}
				ancestor = (ISchemaElement)ancestors.get(i);
				String link = i == ancestors.size()-1 ? null : getElementDocPath(ancestor);
				
				if(link != null){
					writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
				}
				
				writer.print(ancestor.getTagName());
				
				if(link != null){
					writer.print("</A>");
				}
				
				writer.println();
			}
			
			writer.println("</PRE>");
		}
		
		// --- implemented substitution groups or implementations
		if(iElt.getType() == ISchemaElement.SUBSTITUTION_GROUP_TYPE || isAbstract)
		{
			// --- known implementing classes
			writer.println("<DL>");
			writer.println("<DT><B>Known Implementing Elements:</B></DT>");
			writer.println("<DD>");
			ISchemaElement[] allElements = mainSchema.getAllElements();
			int nbImpl = 0;
			for(int i=0; i<allElements.length; i++)
			{
				if(iElt.getMappedClass() == allElements[i].getMappedClass()){
					// skip
					continue;
				}
				
				if(iElt.getMappedClass().isAssignableFrom(allElements[i].getMappedClass()))
				{
					if(nbImpl>0){
						writer.print(", ");
					}
					String link = getElementDocPath(allElements[i]);
					
					if(link != null){
						writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
					}
					
					writer.print(allElements[i].getTagName());
					
					if(link != null){
						writer.print("</A>");
					}
					nbImpl++;
				}
			}
			if(nbImpl == 0){
				writer.print("None.");
			}
			
			writer.println("</DD>");
			writer.println("</DL>");
		}
		
		ISchemaElement[] substitutionGroups = iElt.getSubstitutionGroups();
		if(substitutionGroups != null && substitutionGroups.length > 0)
		{
			writer.println("<DL>");
			writer.println("<DT><B>Substitution Groups:</B></DT>");
			writer.println("<DD>");
			for(int i=0; i<substitutionGroups.length; i++)
			{
				if(i>0){
					writer.print(", ");
				}
				
				String link = getElementDocPath(substitutionGroups[i]);
				
				if(link != null){
					writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
				}
				
				writer.print(substitutionGroups[i].getTagName());
				
				if(link != null){
					writer.print("</A>");
				}
			}
			writer.println("</DD>");
			writer.println("</DL>");
		}
		
		writer.println("<HR>");
		
		// --- details
		String compDetails = iElt.getSchema().getMethod().getElementDetails(locale, iElt.getMappedClass());
		
		if(compDetails == null){
			compDetails = iElt.getSchema().getMethod().getElementTitle(locale, iElt.getMappedClass());
		}
		
		writer.println("<P>");
		writer.println(compDetails);
		
		// --- version, author, ...
		//TODO
		/*
		writer.println("<P>");
		writer.println("<DL>");
		writer.println("<DT><B>Version:</B></DT>");
		writer.println("  <DD>$Version$</DD>");
		writer.println("<DT><B>Author:</B></DT>");
		writer.println("  <DD>Various</DD>");
		writer.println("<DT><B>See Also:</B><DD><A HREF=\"../../../serialized-form.html#javax.servlet.http.HttpServlet\">Serialized Form</A></DL>");
		*/
		
		writer.println("<P>");
		writer.println("<HR>");
		
		if(iElt.getType() == ISchemaElement.ENUMERATION_TYPE)
		{
			// --- write enumeration values
			writer.println("<P>");
			writer.println("<A NAME=\"ENUM_ITEMS\"/>");
			writer.println("<TABLE BORDER=\"1\" WIDTH=\"100%\" CELLPADDING=\"3\" CELLSPACING=\"0\">");
			writer.println("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">");
			writer.println("<TD COLSPAN=2><FONT SIZE=\"+2\">");
			writer.println("<B>Enumeration Items</B></FONT></TD>");
			writer.println("</TR>");
			int nbItems = iElt.getSchema().getMethod().getNbOfEnumItems(iElt.getMappedClass());
			for(int i=0; i<nbItems; i++)
			{
				String itemName = iElt.getSchema().getMethod().getEnumItemName(iElt.getMappedClass(), i);
				String itemDetails = iElt.getSchema().getMethod().getEnumItemDetails(locale, iElt.getMappedClass(), itemName);
				
				if(itemDetails == null){
					itemDetails = iElt.getSchema().getMethod().getEnumItemDetails(locale, iElt.getMappedClass(), itemName);
				}
				
				writer.println("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\">");
				// --- first cell: name
				writer.println("<TD nowrap><CODE><B>"+itemName+"</B></CODE></TD>");
				// --- second cell: details
				writer.println("<TD width=\"100%\">"+itemDetails+"</TD>");
				writer.println("</TR>");
			}
			writer.println("</TABLE>");
		}
		else // ELEMENT or SUBSTITUTION_GROUP
		{
			// --- content
			writer.println("<P>");
			writer.println("<!-- ======== CONTENT ======== -->");
			writer.println("<A NAME=\"CONTENT\"/>");
			for(int i=0; i<fieldAncestors.size(); i++)
			{
				ISchemaElement elt = (ISchemaElement)fieldAncestors.get(fieldAncestors.size()-i-1);
				FieldMapping[] mappedFields = ((ElementMapping)mappings.get(i)).getMappings(FieldMapping.CONTENT);
				if(mappedFields != null && mappedFields.length > 0)
				{
					writer.println("<TABLE BORDER=\"1\" WIDTH=\"100%\" CELLPADDING=\"3\" CELLSPACING=\"0\">");
					writer.println("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">");
					writer.println("<TD COLSPAN=1><FONT SIZE=\"+2\"><B>");
					if(i == 0)
					{
						// --- element attributes
						writer.println("Element Textual Content");
					}
					else
					{
						// --- inherited attributes
						writer.println("Element Textual Content Inherited from ");
						String link = getElementDocPath(elt);
						
						if(link != null){
							writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
						}
						
						writer.print(elt.getTagName());
						
						if(link != null){
							writer.print("</A>");
						}
					}
					writer.println("</B></FONT></TD>");
					writer.println("</TR>");
					
					for(int j=0; j<mappedFields.length; j++)
					{
						FieldMapping field = mappedFields[j];
						String attrDetails = iElt.getSchema().getMethod().getMappingDetails(locale, field);
						
						if(attrDetails == null){
							attrDetails = iElt.getSchema().getMethod().getMappingTitle(locale, field);
						}
						
						// --- first cell: optional|mandatory <type>
						writer.println("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\">");
						writer.println("<TD width=\"100%\">"+attrDetails+"</TD>");
						writer.println("</TR>");
					}
					writer.println("</TABLE>");
					writer.println("<P>");
				}
			}

			// --- attributes
			writer.println("<P>");
			writer.println("<!-- ======== ATTRIBUTES ======== -->");
			writer.println("<A NAME=\"ATTRIBUTES\"/>");
			for(int i=0; i<fieldAncestors.size(); i++)
			{
				ISchemaElement elt = (ISchemaElement)fieldAncestors.get(fieldAncestors.size()-i-1);
				FieldMapping[] mappedFields = ((ElementMapping)mappings.get(i)).getMappings(FieldMapping.ATTRIBUTE);
				if(mappedFields != null && mappedFields.length > 0)
				{
					writer.println("<TABLE BORDER=\"1\" WIDTH=\"100%\" CELLPADDING=\"3\" CELLSPACING=\"0\">");
					writer.println("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">");
					writer.println("<TD COLSPAN=3><FONT SIZE=\"+2\"><B>");
					if(i == 0)
					{
						// --- element attributes
						writer.println("Attributes");
					}
					else
					{
						// --- inherited attributes
						writer.println("Attributes Inherited from ");
						String link = getElementDocPath(elt);
						
						if(link != null){
							writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
						}
						
						writer.print(elt.getTagName());
						
						if(link != null){
							writer.print("</A>");
						}
					}
					writer.println("</B></FONT></TD>");
					writer.println("</TR>");
					
					for(int j=0; j<mappedFields.length; j++)
					{
						FieldMapping field = mappedFields[j];
						String attrDetails = iElt.getSchema().getMethod().getMappingDetails(locale, field);
						
						if(attrDetails == null){
							attrDetails = iElt.getSchema().getMethod().getMappingTitle(locale, field);
						}
						
						// --- first cell: optional|mandatory <type>
						writer.println("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\">");
						writer.println("<TD ALIGN=\"right\" VALIGN=\"top\" WIDTH=\"1%\"><FONT SIZE=\"-1\">");
						writer.println("<CODE>");
						
						if(field.isUseRequired()){
							writer.println("mandatory");
						}else{
							writer.println("optional");
						}
						writer.println("&nbsp;");
						
						if(iElt.getSchema().getMethod().isEnumeration(field.getBaseClass()))
						{
							ISchemaElement enumElt = schemas.getElement(field.getBaseClass());
							if(enumElt != null)
							{
								String link = getElementDocPath(enumElt);
								
								if(link != null){
									writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
								}
								
								writer.print(enumElt.getTagName());
								
								if(link != null){
									writer.print("</A>");
								}
							}
							else
							{
								writer.println(ClassHelper.getSimpleName(field.getBaseClass()));
							}
						}
						else
						{
							writer.println(ClassHelper.getSimpleName(field.getBaseClass()));
						}
						writer.println("</CODE></FONT></TD>");
						
						// --- second cell: name
						writer.println("<TD nowrap><CODE><B>"+field.getName()+"</B></CODE></TD>");
						// --- third cell: details
						writer.println("<TD width=\"100%\">"+attrDetails+"</TD>");
						writer.println("</TR>");
					}
					writer.println("</TABLE>");
					writer.println("<P>");
				}
			}
			
			// --- children Elements
			writer.println("<P>");
			writer.println("<!-- ======== CHILDREN ======== -->");
			writer.println("<A NAME=\"CHILDREN\"/>");
			for(int i=0; i<fieldAncestors.size(); i++)
			{
				ISchemaElement elt = (ISchemaElement)fieldAncestors.get(fieldAncestors.size()-i-1);
				FieldMapping[] mappedFields = ((ElementMapping)mappings.get(i)).getMappings(FieldMapping.CHILD_ELT);
				if(mappedFields != null && mappedFields.length > 0)
				{
					writer.println("<TABLE BORDER=\"1\" WIDTH=\"100%\" CELLPADDING=\"3\" CELLSPACING=\"0\">");
					writer.println("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">");
					writer.println("<TD COLSPAN=3><FONT SIZE=\"+2\"><B>");
					if(i == 0)
					{
						// --- element attributes
						writer.println("Named Children Elements");
					}
					else
					{
						// --- inherited attributes
						writer.println("Named Children Elements Inherited from ");
						String link = getElementDocPath(elt);
						
						if(link != null){
							writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
						}
						
						writer.print(elt.getTagName());
						
						if(link != null){
							writer.print("</A>");
						}
					}
					writer.println("</B></FONT></TD>");
					writer.println("</TR>");
					
					for(int j=0; j<mappedFields.length; j++)
					{
						FieldMapping field = (FieldMapping)mappedFields[j];
						String fieldDetails = iElt.getSchema().getMethod().getMappingDetails(locale, field);
						
						if(fieldDetails == null){
							fieldDetails = iElt.getSchema().getMethod().getMappingTitle(locale, field);
						}
						
						// --- first cell: <range> <type>
						writer.println("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\">");
						writer.println("<TD ALIGN=\"right\" VALIGN=\"top\" WIDTH=\"1%\"><FONT SIZE=\"-1\">");
						writer.println("<CODE>");
						if(field.getMaxOccurs() == 1)
						{
							if(field.isUseRequired()){
								writer.println("mandatory");
							}else{
								writer.println("optional");
							}
						}
						else
						{
							if(field.getMinOccurs() == field.getMaxOccurs()){
								writer.println("extactly "+field.getMaxOccurs());
							}else{
								writer.println(String.valueOf(field.getMinOccurs())+"&nbsp;-&nbsp;"+(field.getMaxOccurs() == Integer.MAX_VALUE ? "unbounded" : String.valueOf(field.getMaxOccurs())));
							}
						}
						writer.println("&nbsp;");
						
						ISchemaElement fieldTypeElt = schemas.getElement(field.getBaseClass());
						if(fieldTypeElt != null)
						{
							String link = getElementDocPath(fieldTypeElt);
							
							if(link != null){
								writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
							}
							
							writer.print(fieldTypeElt.getTagName());
							
							if(link != null){
								writer.print("</A>");
							}
						}
						else
						{
							writer.print(ClassHelper.getSimpleName(field.getBaseClass()));
						}
						
						if(field.getMaxOccurs() > 1){
							writer.print("[]");
						}
						
						writer.println("</CODE></FONT></TD>");
						
						// --- second cell: name
						writer.println("<TD nowrap><CODE><B>"+field.getName()+"</B></CODE></TD>");
						// --- third cell: details
						writer.println("<TD width=\"100%\">"+fieldDetails+"</TD>");
						writer.println("</TR>");
					}
					writer.println("</TABLE>");
					writer.println("<P>");
				}
			}
		}

		// --- collected Elements
		writer.println("<P>");
		writer.println("<!-- ======== CHILDREN ======== -->");
		writer.println("<A NAME=\"CHILDREN\"/>");
		for(int i=0; i<fieldAncestors.size(); i++)
		{
			ISchemaElement elt = (ISchemaElement)fieldAncestors.get(fieldAncestors.size()-i-1);
			FieldMapping[] mappedFields = ((ElementMapping)mappings.get(i)).getMappings(FieldMapping.DIRECT_CHILD_ELT);
			if(mappedFields != null && mappedFields.length > 0)
			{
				writer.println("<TABLE BORDER=\"1\" WIDTH=\"100%\" CELLPADDING=\"3\" CELLSPACING=\"0\">");
				writer.println("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">");
				writer.println("<TD COLSPAN=2><FONT SIZE=\"+2\"><B>");
				if(i == 0)
				{
					// --- element attributes
					writer.println("Direct Children Elements");
				}
				else
				{
					// --- inherited attributes
					writer.println("Direct Children Elements Inherited from ");
					String link = getElementDocPath(elt);
					
					if(link != null){
						writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
					}
					
					writer.print(elt.getTagName());
					
					if(link != null){
						writer.print("</A>");
					}
				}
				writer.println("</B></FONT></TD>");
				writer.println("</TR>");
				
				for(int j=0; j<mappedFields.length; j++)
				{
					FieldMapping field = mappedFields[j];
					String fieldDetails = iElt.getSchema().getMethod().getMappingDetails(locale, field);
					
					if(fieldDetails == null){
						fieldDetails = iElt.getSchema().getMethod().getMappingTitle(locale, field);
					}
					
					// --- first cell: <range> <type>
					writer.println("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\">");
					writer.println("<TD ALIGN=\"right\" VALIGN=\"top\" WIDTH=\"1%\"><FONT SIZE=\"-1\">");
					writer.println("<CODE>");
					if(field.getMaxOccurs() == 1)
					{
						if(field.isUseRequired()){
							writer.println("mandatory");
						} else{
							writer.println("optional");
						}
					}
					else
					{
						if(field.getMinOccurs() == field.getMaxOccurs()){
							writer.println("extactly "+field.getMaxOccurs());
						} else{
							writer.println(String.valueOf(field.getMinOccurs())+"&nbsp;-&nbsp;"+(field.getMaxOccurs() == Integer.MAX_VALUE ? "unbounded" : String.valueOf(field.getMaxOccurs())));
						}
					}
					writer.println("&nbsp;");
					
					ISchemaElement fieldTypeElt = schemas.getElement(field.getBaseClass());
					if(fieldTypeElt != null)
					{
						String link = getElementDocPath(fieldTypeElt);
						
						if(link != null){
							writer.print("<A HREF=\""+toRoot+"/"+link+"\">");
						}
						
						writer.print(fieldTypeElt.getTagName());
						
						if(link != null){
							writer.print("</A>");
						}
					}
					else
					{
						writer.print(ClassHelper.getSimpleName(field.getBaseClass()));
					}
					
					if(field.getMaxOccurs() > 1){
						writer.print("[]");
					}
					
					writer.println("</CODE></FONT></TD>");
					
					// --- second cell: details
					writer.println("<TD width=\"100%\">"+fieldDetails+"</TD>");
					writer.println("</TR>");
				}
				writer.println("</TABLE>");
				writer.println("<P>");
			}
		}
		
		writer.println("</BODY>");
		writer.println("</HTML>");
	
		writer.close();
	}

}
