/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.ow2.opensuit.xmlmap.impl.IXmlModelMessage;
import org.ow2.opensuit.xmlmap.impl.ObjInfo;
import org.ow2.opensuit.xmlmap.impl.ObjInfoHolder;
import org.ow2.opensuit.xmlmap.mapping.ElementMapping;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.schema.ISchema;
import org.ow2.opensuit.xmlmap.schema.ISchemaElement;
import org.ow2.opensuit.xmlmap.schema.SchemasManager;
import org.ow2.opensuit.xmlmap.utils.XmlHelper;
import org.ow2.opensuit.xmlmap.utils.XsdHelper;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;



/**
 * The base class for instantiating an object model that conform to XML mapping rules.
 */
public class XmlSerializer extends ObjInfoHolder
{
	private Document doc;
	private SchemasManager manager;
	private Object root;
	

	/**
	 * Returns the root instantiated object.
	 * @return
	 */
	public Document getDocument()
	{
		return doc;
	}
	/**
	 * Instantiates a declarative model that conforms to implicit XML Mapping rules.
	 * @param iInput The XML input stream
	 * @return The root Object object of the model.
	 * @throws ParserConfigurationException 
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 * @throws SAXException 
	 */
	public final static XmlSerializer serialize(SchemasManager iManager, Object iRoot) throws ParserConfigurationException
	{
		XmlSerializer serializer = new XmlSerializer(iManager, iRoot);
		serializer.serialize();
		return serializer;
	}
	public final static XmlSerializer serialize(SchemasManager iManager, Object iRoot, OutputStream output) throws ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException, UnsupportedEncodingException
	{
		XmlSerializer serializer = serialize(iManager, iRoot);
		XmlHelper.serializeXml(serializer.getDocument(), output);
		return serializer;
	}
	// ===============================================================================
	
	public XmlSerializer(SchemasManager iManager, Object iRoot) throws ParserConfigurationException
    {
		DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		doc = docBuilder.newDocument();
		manager = iManager;
		root = iRoot;
    }

	private void serialize()
	{
		ObjInfo rootInfo = new ObjInfo(null, root);
		addInfo(rootInfo);
		serializeObj(doc, rootInfo);
		
		// --- add namespaces on root element
		/*
			xmlns:base="org.ow2.opensuit.xml.base"
			xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
			xsi:schemaLocation="org.ow2.opensuit.xml.base base.xsd"
		 */
		Element element = doc.getDocumentElement();
//		element.setAttributeNS("xmlns", "xsi", "http://www.w3.org/2001/XMLSchema-instance");
		element.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
		
		ISchema[] imports = manager.getAllSchemas();
		StringBuffer schemaLocation = new StringBuffer();
		for(int i=0; i<imports.length; i++)
		{
//			element.setAttributeNS("xmlns", imports[i].getName(), imports[i].getRootPackage());
			element.setAttribute("xmlns:"+imports[i].getName(), imports[i].getNamespace());
			schemaLocation.append(imports[i].getNamespace());
			schemaLocation.append(" ");
			schemaLocation.append(imports[i].getLocator());
			schemaLocation.append(" ");
		}
		
//		element.setAttributeNS("xsi", "schemaLocation", schemaLocation.toString());
		element.setAttribute("xsi:schemaLocation", schemaLocation.toString());
	}
	private void serializeObj(Node iParent, ObjInfo iObjInfo)
	{
		ISchemaElement schemaElt = manager.getElement(iObjInfo.getObject().getClass());
//		ISchemaElement schemaElt = iObjInfo.getSchemaElement();
		if(schemaElt == null)
		{
			iObjInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "Object not found in any declared schema.");
			return;
		}
		
		Element element = doc.createElement(schemaElt.getSchema().getName()+":"+schemaElt.getTagName());
		iParent.appendChild(element);
		// --- extract XmlMappings
		ElementMapping mappings = (ElementMapping)schemaElt.getSchema().getMethod().getElementMappings(iObjInfo.getObject().getClass(), true);

		// =========================================================
		// === Process Content Mappings
		// =========================================================
		FieldMapping[] contentMappings = mappings.getMappings(FieldMapping.CONTENT);
		if(contentMappings.length > 0)
		{
			for(int i=0; i<contentMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)contentMappings[i];
				
				fieldMapping.getField().setAccessible(true);
                try
                {
	                String content = (String)fieldMapping.getField().get(iObjInfo.getObject());
					// --- set content
	                if(content != null)
	                {
	                	CDATASection cdata = doc.createCDATASection(content);
	                	element.appendChild(cdata);
	                }
                }
                catch(Exception e)
                {
					iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content field '"+fieldMapping.getField().getName()+"' could not be retreived: "+e.getMessage());
					continue;
                }
			}
		}
		// =========================================================
		// === Process Attributes
		// =========================================================
		FieldMapping[] attrMappings = mappings.getMappings(FieldMapping.ATTRIBUTE);
		for(int i=0; i<attrMappings.length; i++)
		{
			FieldMapping fieldMapping = (FieldMapping)attrMappings[i];

			// --- get the value
			fieldMapping.getField().setAccessible(true);
			Object val = null;
            try
            {
	            val = fieldMapping.getField().get(iObjInfo.getObject());
            }
            catch(Exception e)
            {
				iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Attribute field '"+fieldMapping.getField().getName()+"' could not be retreived: "+e.getMessage());
				continue;
            }
			
			if(val == null)
			{
				if(fieldMapping.isUseRequired()){
					iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Attribute '"+fieldMapping.getName()+"' required.");
				}
			}
			else
			{
				// --- try to set the field value
				// --- 1: cast to the expected type
				String simpleValue = XsdHelper.formatSimpleType(val, fieldMapping.getBaseClass());
				if(simpleValue != null)
				{
					element.setAttribute(fieldMapping.getName(), simpleValue);
				}
				else if(schemaElt.getSchema().getMethod().isEnumeration(fieldMapping.getBaseClass()))
				{
					// --- field is an enumerated type
					String enumValName = schemaElt.getSchema().getMethod().getEnumItemName(val);
					element.setAttribute(fieldMapping.getName(), enumValName);
				}
			}
		}
		
		// =========================================================
		// === Process Children Elements
		// =========================================================
		// --- 2: set fields mapped to single nodes
		FieldMapping[] childMappings = mappings.getMappings(FieldMapping.CHILD_ELT);
		for(int i=0; i<childMappings.length; i++)
		{
			FieldMapping fieldMapping = (FieldMapping)childMappings[i];
			// --- get the value
			fieldMapping.getField().setAccessible(true);
			Object val = null;
            try
            {
	            val = fieldMapping.getField().get(iObjInfo.getObject());
            }
            catch(Exception e)
            {
				iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node field '"+fieldMapping.getField().getName()+"' could not be retreived: "+e.getMessage());
				continue;
            }
			
			if(val == null)
			{
				if(fieldMapping.isUseRequired()){
					iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' required.");
				}
			}
			else
			{
				Element childNode = doc.createElement(fieldMapping.getName());
				element.appendChild(childNode);
				
				String simpleValue = XsdHelper.formatSimpleType(val, fieldMapping.getBaseClass());
				if(simpleValue != null)
				{
					childNode.appendChild(doc.createTextNode(simpleValue));
				}
				else if(schemaElt.getSchema().getMethod().isEnumeration(fieldMapping.getBaseClass()))
				{
					// --- field is an enumerated type
					String enumValName = schemaElt.getSchema().getMethod().getEnumItemName(val);
					childNode.appendChild(doc.createTextNode(enumValName));
				}
				else
				{
					// --- complex type
					if(fieldMapping.getField().getType().isArray())
					{
						// --- array type
						int len = Array.getLength(val);
						if(len < fieldMapping.getMinOccurs())
						{
							iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at least "+fieldMapping.getMinOccurs()+" children.");
						}
						else if(len > fieldMapping.getMaxOccurs())
						{
							iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at most "+fieldMapping.getMaxOccurs()+" children.");
						}
						
						for(int j=0; j<len; j++)
						{
							Object v = (Object)Array.get(val, j);
							
							if(v == null){
								continue;
							}
							
							ObjInfo childInfo = new ObjInfo(iObjInfo, v);
							addInfo(childInfo);
							serializeObj(childNode, childInfo);
						}
					}
					else
					{
						// --- simple type
						ObjInfo childInfo = new ObjInfo(iObjInfo, (Object)val);
						addInfo(childInfo);
						serializeObj(childNode, childInfo);
					}
				}
			}
		}
		
		// =========================================================
		// === Process Collected Elements
		// =========================================================
		FieldMapping[] collectedMappings = mappings.getMappings(FieldMapping.DIRECT_CHILD_ELT);
		for(int i=0; i<collectedMappings.length; i++)
		{
			FieldMapping fieldMapping = (FieldMapping)collectedMappings[i];
			// --- get the value
			fieldMapping.getField().setAccessible(true);
			Object val = null;
            try
            {
	            val = fieldMapping.getField().get(iObjInfo.getObject());
            }
            catch(Exception e)
            {
				iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Collected node '"+fieldMapping.getField().getName()+"' could not be retreived: "+e.getMessage());
				continue;
            }
			
			if(val == null)
			{
				if(fieldMapping.isUseRequired()){
					iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Collected node '"+fieldMapping.getName()+"' required.");
				}
			}
			else
			{
				if(fieldMapping.getField().getType().isArray())
				{
					// --- array type
					int len = Array.getLength(val);
					if(len < fieldMapping.getMinOccurs())
					{
						iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Collected node '"+fieldMapping.getName()+"' expects at least "+fieldMapping.getMinOccurs()+" children.");
					}
					else if(len > fieldMapping.getMaxOccurs())
					{
						iObjInfo.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Collected node '"+fieldMapping.getName()+"' expects at most "+fieldMapping.getMaxOccurs()+" children.");
					}
					
					for(int j=0; j<len; j++)
					{
						Object v = (Object)Array.get(val, j);
						ObjInfo childInfo = new ObjInfo(iObjInfo, v);
						addInfo(childInfo);
						serializeObj(element, childInfo);
					}
				}
				else
				{
					// --- simple type
					ObjInfo childInfo = new ObjInfo(iObjInfo, (Object)val);
					addInfo(childInfo);
					serializeObj(element, childInfo);
				}
			}
		}
	}
}
