/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap.impl;

import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;


/**
 * The base class for instantiating an object model that conform to XML mapping rules.
 */
public abstract class ObjInfoHolder
{
	private List<ObjInfo> elements = new ArrayList<ObjInfo>();
	private List<URL> files = new ArrayList<URL>();
	private static final URL[] URL_ARRAY_TYPE = {};
	
	// ===============================================================================
	// === Internal info <--> obj mapping methods
	// ===============================================================================
	public List<ObjInfo> getAllObjInfo()
	{
		return this.elements;
	}
	public ObjInfo getInfoFromObj(Object iObj)
	{
		for(int i=0; i<this.elements.size(); i++)
		{
			if(iObj == ((ObjInfo)this.elements.get(i)).getObject()){
				return (ObjInfo)this.elements.get(i);
			}
		}
		return null;
	}
	protected void addInfo(ObjInfo info)
	{
	    this.elements.add(info);
		if(info.getXmlFile() != null && ! this.files.contains(info.getXmlFile())){
		    this.files.add(info.getXmlFile());
		}
	}
	protected void removeAllInfo()
	{
	    this.elements.clear();
	    this.files.clear();
	}
	public URL[] getAllFiles()
	{
		return (URL[]) this.files.toArray(URL_ARRAY_TYPE);
	}
	// ===============================================================================
	// === Messages management methods
	// ===============================================================================
	private int countMessages(int treshold)
	{
		int nbErrs = 0;
		for(int i=0; i < this.elements.size(); i++)
		{
			ObjInfo oi = (ObjInfo) this.elements.get(i);
			nbErrs += oi.countMessages(treshold);
		}
		return nbErrs;
	}

	/**
	 * Dumps errors, warnings and informations in the given printer.
	 * @param iPrinter A printer object to write errors to.
	 */
	public void dumpMessages(int treshold, PrintWriter iPrinter)
	{
		for(int i=0; i < this.elements.size(); i++){
			((ObjInfo) this.elements.get(i)).dumpMessages(treshold, iPrinter);
		}
		
		iPrinter.flush();
	}
	/**
	 * Determines whether this portal has errors (instanciation/validation) or not.
	 * @return <code>true</code> if the portal has errors, <code>false</code> otherwise.
	 */
	public boolean hasMessages(int treshold)
	{
		return countMessages(treshold) > 0;
	}
	public IXmlModelMessage[] getAllMessages()
	{
		List<IXmlModelMessage> retList = new ArrayList<IXmlModelMessage>();
		for(int i=0; i < this.elements.size(); i++)
		{
			ObjInfo oi = (ObjInfo) this.elements.get(i);
			List<IXmlModelMessage> msgs = oi.getMessages(IXmlModelMessage.VALID_INFO);
			
			if(msgs != null){
				retList.addAll(msgs);
			}
		}
		IXmlModelMessage[] ret = new IXmlModelMessage[retList.size()];
		retList.toArray(ret);
		return ret;
	}
	public IXmlModelMessage[] getAllElementMessages(Object iXmlElement)
	{
		ObjInfo oi = getInfoFromObj(iXmlElement);
		
		if(oi == null){
			return null;
		}
		
		List<IXmlModelMessage> msgs = oi.getMessages(IXmlModelMessage.VALID_INFO);
		
		if(msgs == null){
			return null;
		}
		
		IXmlModelMessage[] ret = new IXmlModelMessage[msgs.size()];
		msgs.toArray(ret);
		return ret;
	}
	public IXmlModelMessage[] getGlobalElementMessages(Object iXmlElement)
	{
		ObjInfo oi = getInfoFromObj(iXmlElement);
		
		if(oi == null){
			return null;
		}
		
		List<IXmlModelMessage> msgs = oi.getMessages(IXmlModelMessage.VALID_INFO);
		
		if(msgs == null){
			return null;
		}
		
		List<IXmlModelMessage> retList = new ArrayList<IXmlModelMessage>();
		for(int i=0; i<msgs.size(); i++)
		{
			IXmlModelMessage msg = (IXmlModelMessage)msgs.get(i);
			
			if(msg.getMappingName() == null){
				retList.add(msg);
			}
		}
		
		if(retList.isEmpty()){
			return null;
		}
		
		IXmlModelMessage[] ret = new IXmlModelMessage[retList.size()];
		retList.toArray(ret);
		return ret;
	}
	public IXmlModelMessage[] getMappingElementMessages(Object iXmlElement, String iMappingName)
	{
		ObjInfo oi = getInfoFromObj(iXmlElement);
		
		if(oi == null){
			return null;
		}
		
		List<IXmlModelMessage> msgs = oi.getMessages(IXmlModelMessage.VALID_INFO);
		
		if(msgs == null){
			return null;
		}
		
		List<IXmlModelMessage> retList = new ArrayList<IXmlModelMessage>();
		for(int i=0; i<msgs.size(); i++)
		{
			IXmlModelMessage msg = (IXmlModelMessage)msgs.get(i);
			
			if(iMappingName.equals(msg.getMappingName())){
				retList.add(msg);
			}
		}
		
		if(retList.isEmpty()){
			return null;
		}
		
		IXmlModelMessage[] ret = new IXmlModelMessage[retList.size()];
		retList.toArray(ret);
		return ret;
	}
}
