/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap.mapping;

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.ow2.opensuit.xmlmap.interfaces.IContentAssist;


public abstract class IMappingMethod
{
	private static final MappingError[] MAPPING_ERROR_ARRAY_TYPE = {};
	
	private static Map<String ,ElementMapping> class2Mappings = new HashMap<String ,ElementMapping>();
	
	// =====================================================================================
	// === Type Methods
	// =====================================================================================
	public abstract boolean isSimpleType(Class<?> c);
	
	public abstract boolean isXmlElement(Class<?> c);
	
	public abstract boolean isSubstitutionGroup(Class<?> c);
	
	public abstract boolean isEnumeration(Class<?> c);
	
	public abstract boolean isImportable(Class<?> c);
	
	protected abstract void populateMappings(List<?> mappedFields, List<?> mappingErrors, Class<?> iClass, Class<?> iExcludeFrom);

	// =====================================================================================
	// === SimpleType Methods
	// =====================================================================================
	public abstract Class<?> getBaseSimpleType(Class<?> c);
	
	// =====================================================================================
	// === Element Methods
	// =====================================================================================
	/**
	 * Returns the Xml Mappings for the given class.
	 * This method allows to retreive all mappings from the whole class inheritance,
	 * or from the class only.
	 * @param iClass The class to extract Xml Mappings from.
	 * @param iGetInherited if <code>true</code>, the method will return mappings declared anywhere in the class inheritance graph,
	 * Otherwise will return mappings declared in this class only.
	 * @return An object that contains Xml Mappings
	 */
	public final ElementMapping getElementMappings(Class<?> iClass, boolean iGetInherited)
	{
		String key = iClass.getName();
		
		if(iGetInherited){
			key = "(full)@"+key;
		}
		
		ElementMapping mappings = (ElementMapping)class2Mappings.get(key);
		if(mappings == null)
		{
			// --- mappings not retrieved from cached: have to compute
			List mappedFields = new ArrayList();
			List<MappingError> mappingErrors = new ArrayList<MappingError>();
			
			// --- 1: check that the given class has a default constructor that 
			// --- is accessible (only if implements IXmlObject and non-asbtract)
			Constructor<?> cs = null;
			if(isXmlElement(iClass) && (iClass.getModifiers() & Modifier.ABSTRACT) == 0)
			{
				try
		        {
			        cs = iClass.getConstructor(new Class[]{});
			        if(cs == null)
			        {
		        		mappingErrors.add(new MappingError("Class '"+iClass.getName()+"' has no default constructor."));
			        }
			        //else
			        //{
		// no	        	if(!cs.isAccessible())
		//	        		mappingErrors.addElement(new XmlMappingException("Class '"+iClass.getName()+"' default constructor is not accessible."));
			        //}
		        }
		        catch(SecurityException e)
		        {
		    		mappingErrors.add(new MappingError("SecurityError while retrieveing '"+iClass.getName()+"' default constructor: "+e.getMessage()));
		        }
		        catch(NoSuchMethodException e)
		        {
		    		mappingErrors.add(new MappingError("Class '"+iClass.getName()+"' has no default constructor."));
		        }
			}
			
	        Class<?> excludeFromClass = iGetInherited ? Object.class : iClass.getSuperclass();
	        populateMappings(mappedFields, mappingErrors, iClass, excludeFromClass);
	        
			MappingError[] errors = (MappingError[])mappingErrors.toArray(MAPPING_ERROR_ARRAY_TYPE);
			mappings = new ElementMapping(iClass, cs, mappedFields, errors);
			
			class2Mappings.put(key, mappings);
		}
		return mappings;
	}
	// =====================================================================================
	// === Enumeration Methods
	// =====================================================================================
	public abstract int getNbOfEnumItems(Class<?> iEnumClass);
	/**
	 * Returns an enumeration element from its name.
	 * @param iName The enumeration element name.
	 * @return The enumeration element value.
	 */
	public abstract Object getEnumItem(Class<?> iEnumClass, String iName);

	/**
	 * Returns an enumeration element from its index.
	 * @param iIndex The enumeration element index.
	 * @return The enumeration element value.
	 */
	public abstract Object getEnumItem(Class<?> iEnumClass, int iIndex);

	/**
	 * Returns an enumeration element name from its index.
	 * @param iIndex The enumeration element index.
	 * @return The enumeration element name.
	 */
	public abstract String getEnumItemName(Class<?> iEnumClass, int iIndex);

	public abstract String getEnumItemName(Object iEnumValue);
	
	// =====================================================================================
	// === Documentation Methods
	// =====================================================================================
	public abstract String getInstanceName(Locale iLocale, Object iElement);

	public abstract String getElementTitle(Locale iLocale, Class<?> iElementClass);

	public abstract String getElementDetails(Locale iLocale, Class<?> iElementClass);

	public abstract String getMappingTitle(Locale iLocale, FieldMapping mapping);

	public abstract String getMappingDetails(Locale iLocale, FieldMapping mapping);

	public abstract String getEnumItemDetails(Locale iLocale, Class<?> iEnumClass, String iName);
	
	// =====================================================================================
	// === Content Assist Methods
	// =====================================================================================
	public abstract IContentAssist getContentAssist(Object obj, FieldMapping mapping);
}
