/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap.schema;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.opensuit.xmlmap.impl.SchemaImpl;


/**
 * Main object to instanciate and manage schemas.
 */
public class SchemasManager
{
	private static final ISchema[] SCHEMAS_ARRAY_TYPE = {};
	private static final ISchemaElement[] ELTS_ARRAY_TYPE = {};
	
	private List<ISchema> schemas = new ArrayList<ISchema>();
	private Map<Class<?>,ISchemaElement[]> interface2Implementors = new HashMap<Class<?>,ISchemaElement[]>();
	
	private ClassLoader classLoader = SchemaImpl.class.getClassLoader();
	
	public ClassLoader getClassLoader()
	{
		return classLoader;
	}
	public void setClassLoader(ClassLoader classLoader)
	{
		this.classLoader = classLoader;
	}

	/**
	 * Declares a schema.
	 * @param iName The schema prefix
	 * @param iLocator The schema locator (xsd file location)
	 * @param iPackage The root schema package
	 * @return
	 */
	public ISchema declareSchema(String iName, String iLocator, String iPackage)
	{
		ISchema schema = getSchema(iPackage);
		if(schema == null)
		{
			schema = new SchemaImpl(this, iName, iLocator, iPackage);
			schemas.add(schema);
		}
		return schema;
	}
	/**
	 * Returns a schema by its root package.
	 * @param iPackage The schema root package
	 * @return The matching schema.
	 */
	public ISchema getSchema(String iPackage)
	{
		for(int i=0; i<schemas.size(); i++)
		{
			ISchema schema = (ISchema)schemas.get(i);
			
			if(iPackage.equals(schema.getRootPackage())){
				return schema;
			}
		}
		return null;
	}
	/**
	 * Builds all declared schemas from the given source code directory.
	 * @param iSourceRootDir The source code directory.
	 * @throws Exception
	 */
	/*
	public void buildAllFromSource(String iSourceRootDir) throws Exception
	{
		for(int i=0; i<schemas.size(); i++)
		{
			SchemaImpl schema = (SchemaImpl)schemas.get(i);
			schema.buildFromSource(iSourceRootDir);
		}
	}
	*/
	public void buildAllFromClassPath() throws Exception
	{
		for(int i=0; i<schemas.size(); i++)
		{
			SchemaImpl schema = (SchemaImpl)schemas.get(i);
			schema.buildFromClassPath();
		}
	}
	/**
	 * Returns all declared schemas
	 * @return
	 */
	public ISchema[] getAllSchemas()
	{
		return (ISchema[])schemas.toArray(SCHEMAS_ARRAY_TYPE);
	}
	/**
	 * Returns the schema element describing the given class, if found in any
	 * declared schema.
	 * @param iClass An object class
	 * @return The corresponding schema element
	 */
	public ISchemaElement getElement(Class<?> iClass)
	{
		SchemaImpl schema = (SchemaImpl)getDeclaringSchema(iClass);
		
		if(schema == null){
			return null;
		}
		
		return schema.getElement(iClass);
	}
	/**
	 * Returns all schema elements that are compatible with the given class
	 * @param iSubstitutionGroup
	 * @return
	 */
	public ISchemaElement[] getAllImplementors(Class<?> iSubstitutionGroup)
	{
		ISchemaElement[] implementors = (ISchemaElement[])interface2Implementors.get(iSubstitutionGroup);
		if(implementors == null)
		{
			// --- look for all non abstract classes that implement the given interface
			List<ISchemaElement> implementorsVect = new ArrayList<ISchemaElement>();
			
			for(int i=0; i<schemas.size(); i++)
			{
				ISchema schema = (ISchema)schemas.get(i);
				ISchemaElement[] elements = schema.getAllElements();
				for(int j=0; j<elements.length; j++)
				{
					if(!elements[j].getMappedClass().isInterface() &&  (elements[j].getMappedClass().getModifiers() & Modifier.ABSTRACT) == 0 && iSubstitutionGroup.isAssignableFrom(elements[j].getMappedClass())){
						implementorsVect.add(elements[j]);
					}
				}
			}
			
			implementors = (ISchemaElement[])implementorsVect.toArray(ELTS_ARRAY_TYPE);
			interface2Implementors.put(iSubstitutionGroup, implementors);
		}
		return implementors;
	}
	/**
	 * Returns the schema the given class belongs to (if any)
	 * @param iClass
	 * @return
	 */
	public ISchema getDeclaringSchema(Class<?> iClass)
	{
		String classname = iClass.getName();
		int lastDotIdx = classname.lastIndexOf('.');
		
		if(lastDotIdx < 0){
			return null;
		}
		
		String pack = classname.substring(0, lastDotIdx);
		
		// --- look in imports
		if(schemas != null)
		{
			for(int i=0; i<schemas.size(); i++)
			{
				ISchema schema = (ISchema)schemas.get(i);
//				if(imp._package.equals(pack))
				
				if(pack.startsWith(schema.getRootPackage())){
					return schema;
				}
			}
		}
		return null;
	}
	
}
