/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap.utils.minidom;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;



/**
 * The base class for instantiating an object model that conform to XML mapping rules.
 */
public class Parser
{
	public final static XmlElement parse(URL document) throws IOException, SAXException, ParserConfigurationException, FactoryConfigurationError
	{
		SAXParserFactory parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
		SAXParser parser = parserFactory.newSAXParser();
		
		// --- now parse
		InputSource input = new InputSource(document.toExternalForm());
		SAXHandler hdl = new SAXHandler();
//		InputStream in = document.openStream();
		parser.parse(input, hdl);
//		in.close();
		
		return hdl.getDocumentElement().getFirstChild();
	}

	// ======================================================================
	// === SAX Handler
	// ======================================================================
	private static class SAXHandler extends DefaultHandler
	{
		private Locator locator = null;
		private List<XmlElement> ancestors = new ArrayList<XmlElement>();
		
		private void startElement(XmlElement elt)
		{
			XmlElement parent = getCurElement();
			
			if(parent != null){
				parent.addChild(elt);
			}
			
			ancestors.add(elt);
		}
		private void endElement()
		{
			if(ancestors.size() > 1){
				ancestors.remove(ancestors.size()-1);
			}
		}
		private XmlElement getDocumentElement()
		{
			return (XmlElement)ancestors.get(0);
		}
		private XmlElement getCurElement()
		{
			if(ancestors.size() == 0){
				return null;
			}
			
			return (XmlElement)ancestors.get(ancestors.size()-1);
		}
		private void trace(String message)
		{
			/*
			if(locator == null)
				System.out.println("[?] "+message);
			else
				System.out.println("[l"+locator.getLineNumber()+" c"+locator.getColumnNumber()+"] "+message);
			*/
		}
		
		// ====================================================================
		// === SAX Handler Methods
		// ====================================================================
		public void setDocumentLocator(Locator locator)
		{
			trace("document locator: "+locator.getPublicId()+"/"+locator.getSystemId());
			this.locator = locator;
		}
		public void startDocument() throws SAXException
		{
			trace("startDocument()");
			
			XmlElement elt = new XmlElement(null, null, null);
			startElement(elt);
		}
		public void endDocument() throws SAXException
		{
			trace("endDocument()");
		}
		public void startElement(String uri, String localName, String name, Attributes attributes) throws SAXException
		{
			trace("startElement('"+uri+"', '"+localName+"', '"+name+"')");
			
			XmlElement elt = new XmlElement(uri, localName, name);
			
			if(locator != null){
				elt.setLocation(locator.getLineNumber(), locator.getColumnNumber());
			}
			
			if(attributes != null)
			{
				for(int i=0; i<attributes.getLength(); i++)
				{
					XmlAttribute attr = new XmlAttribute(attributes.getURI(i), attributes.getLocalName(i), attributes.getQName(i), attributes.getValue(i));
					elt.addAttribute(attr);
				}
			}
			startElement(elt);
		}
		public void characters(char[] chars, int start, int length) throws SAXException
		{
//			trace("characters: "+new String(chars, start, length));
//			trace("characters: "+length);
			getCurElement().appendContent(chars, start, length);
		}
		public void endElement(String uri, String localName, String name) throws SAXException
		{
			trace("endElement('"+uri+"', '"+localName+"', '"+name+"')");
			endElement();
		}
		
		public void startPrefixMapping(String prefix, String uri) throws SAXException
		{
			trace("startPrefixMapping('"+prefix+"', '"+uri+"')");
		}
		public void endPrefixMapping(String prefix) throws SAXException
		{
			trace("endPrefixMapping('"+prefix+"')");
		}
		public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
		{
			trace("ignorableWhitespace(): "+length);
		}
		public void processingInstruction(String target, String data) throws SAXException
		{
			trace("processingInstruction('"+target+"', '"+data+"')");
		}
		public void skippedEntity(String name) throws SAXException
		{
			trace("skippedEntity('"+name+"')");
		}
	}

}
